# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

from linien_common.common import pack


class Parameter:
    """Represents a single parameter and is used by `Parameters`."""

    def __init__(
        self,
        min_=None,
        max_=None,
        start=None,
        wrap=False,
        sync=True,
        collapsed_sync=True,
    ):
        self.min = min_
        self.max = max_
        self.wrap = wrap
        self._value = start
        self._start = start
        self._listeners = set()
        self._collapsed_sync = collapsed_sync
        self.exposed_can_be_cached = sync

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        # check bounds
        if self.min is not None and value < self.min:
            value = self.min if not self.wrap else self.max
        if self.max is not None and value > self.max:
            value = self.max if not self.wrap else self.min

        self._value = value

        # we copy it because a listener could remove a listener --> this would
        # cause an error in this loop
        for listener in self._listeners.copy():
            listener(value)

    def on_change(self, function, call_listener_with_first_value=True):
        self._listeners.add(function)

        if call_listener_with_first_value:
            if self._value is not None:
                function(self._value)

    def remove_listener(self, function):
        if function in self._listeners:
            self._listeners.remove(function)

    def exposed_reset(self):
        self.value = self._start

    def register_remote_listener(self, remote_uuid):
        pass


class BaseParameters:
    """Represents a set of parameters. In an actual program, it should be
    sub-classed like this:

        class MyParameters(BaseParameters):
            def __init__(self):
                self.param1 = Parameter(min_=12, max_=24)

    Parameters can be changed like this:

        p = MyParameters(...)
        p.param1.value = 123

    You can register callback functions like this:

        def on_change(value):
            # do something

        p.param1.on_change(on_change)
    """

    def __init__(self):
        self._remote_listener_queue = {}
        self._remote_listener_callbacks = {}

    def get_all_parameters(self):
        for name, element in self.__dict__.items():
            if isinstance(element, Parameter):
                yield name, element

    def init_parameter_sync(self, uuid):
        """To be called by a remote client: Yields all parameters as well
        as their values and if the parameters are suited to be cached registers
        a listener that pushes changes of these parameters to the client."""
        for name, element in self.get_all_parameters():
            yield name, element, element.value, element.exposed_can_be_cached
            if element.exposed_can_be_cached:
                self.register_remote_listener(uuid, name)

    def register_remote_listener(self, uuid, param_name):
        self._remote_listener_queue.setdefault(uuid, [])
        self._remote_listener_callbacks.setdefault(uuid, [])

        def on_change(value, uuid=uuid, param_name=param_name):
            if uuid in self._remote_listener_queue:
                self._remote_listener_queue[uuid].append((param_name, value))

        param = getattr(self, param_name)
        param.on_change(on_change)

        self._remote_listener_callbacks[uuid].append((param, on_change))

    def unregister_remote_listeners(self, uuid):
        for param, callback in self._remote_listener_callbacks[uuid]:
            param.remove_listener(callback)

        del self._remote_listener_queue[uuid]
        del self._remote_listener_callbacks[uuid]

    def get_listener_queue(self, uuid):
        queue = self._remote_listener_queue.get(uuid, [])
        self._remote_listener_queue[uuid] = []

        # filter out multiple values for collapsible parameters
        already_has_value = []
        for idx in reversed(range(len(queue))):
            param_name, value = queue[idx]
            if self._get_param(param_name)._collapsed_sync:
                if param_name in already_has_value:
                    del queue[idx]
                else:
                    already_has_value.append(param_name)

        return pack(queue)

    def __iter__(self):
        for name, param in self.get_all_parameters():
            yield name, param.value

    def _get_param(self, param_name):
        param = getattr(self, param_name)
        assert isinstance(param, Parameter)
        return param



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import pickle
import traceback

from linien_common.common import check_plot_data, combine_error_signal, get_lock_point
from linien_server.autolock.algorithm_selection import AutolockAlgorithmSelector
from linien_server.autolock.fast import FastAutolock
from linien_server.autolock.robust import RobustAutolock


class Autolock:
    def __init__(self, control, parameters):
        self.control = control
        self.parameters = parameters

        self.first_error_signal = None
        self.first_error_signal_rolled = None
        self.parameters.autolock_running.value = False
        self.parameters.autolock_retrying.value = False

        self.should_watch_lock = False
        self._data_listener_added = False

        self.reset_properties()

        self.algorithm = None

    def reset_properties(self):
        # we check each parameter before setting it because otherwise
        # this may crash the client if called very often (e.g.if the
        # autolock continuously fails)
        if self.parameters.autolock_failed.value:
            self.parameters.autolock_failed.value = False
        if self.parameters.autolock_locked.value:
            self.parameters.autolock_locked.value = False
        if self.parameters.autolock_watching.value:
            self.parameters.autolock_watching.value = False

    def run(
        self,
        x0,
        x1,
        spectrum,
        should_watch_lock=False,
        auto_offset=True,
        additional_spectra=None,
    ):
        """Starts the autolock.

        If `should_watch_lock` is specified, the autolock continuously monitors
        the control and error signals after the lock was successful and tries to
        relock automatically using the spectrum that was recorded in the first
        run of the lock.
        """
        self.parameters.autolock_running.value = True
        self.parameters.autolock_preparing.value = True
        self.parameters.autolock_percentage.value = 0
        self.parameters.fetch_additional_signals.value = False
        self.x0, self.x1 = int(x0), int(x1)
        self.should_watch_lock = should_watch_lock

        # collect parameters that should be restored after stopping the lock
        self.parameters.autolock_initial_sweep_amplitude.value = (
            self.parameters.sweep_amplitude.value
        )

        self.additional_spectra = additional_spectra or []

        (
            self.first_error_signal,
            self.first_error_signal_rolled,
            self.line_width,
            self.peak_idxs,
        ) = self.record_first_error_signal(spectrum, auto_offset)

        try:
            self.autolock_mode_detector = AutolockAlgorithmSelector(
                self.parameters.autolock_mode_preference.value,
                spectrum,
                additional_spectra,
                self.line_width,
            )

            if self.autolock_mode_detector.done:
                self.start_autolock(self.autolock_mode_detector.mode)

        except Exception:
            # this may happen if `additional_spectra` contain uncorrelated data
            # then either autolock algorithm selector or `start_autolock` may
            # raise a spectrum uncorrelated exception
            traceback.print_exc()
            self.parameters.autolock_failed.value = True
            return self.exposed_stop()

        self.add_data_listener()

    def start_autolock(self, mode):
        print("start autolock with mode", mode)
        self.parameters.autolock_mode.value = mode

        self.algorithm = [None, RobustAutolock, FastAutolock][mode](
            self.control,
            self.parameters,
            self.first_error_signal,
            self.first_error_signal_rolled,
            self.peak_idxs[0],
            self.peak_idxs[1],
            additional_spectra=self.additional_spectra,
        )

    def add_data_listener(self):
        if not self._data_listener_added:
            self._data_listener_added = True
            self.parameters.to_plot.on_change(
                self.react_to_new_spectrum, call_listener_with_first_value=False
            )

    def remove_data_listener(self):
        self._data_listener_added = False
        self.parameters.to_plot.remove_listener(self.react_to_new_spectrum)

    def react_to_new_spectrum(self, plot_data):
        """React to new spectrum data.

        If this is executed for the first time, a reference spectrum is
        recorded.

        If the autolock is approaching the desired line, a correlation
        function of the spectrum with the reference spectrum is calculated
        and the laser current is adapted such that the targeted line is centered.

        After this procedure is done, the real lock is turned on and after some
        time the lock is verified.

        If automatic relocking is desired, the control and error signals are
        continuously monitored after locking.
        """
        if self.parameters.pause_acquisition.value:
            return

        if plot_data is None or not self.parameters.autolock_running.value:
            return

        plot_data = pickle.loads(plot_data)
        if plot_data is None:
            return

        is_locked = self.parameters.lock.value

        # check that `plot_data` contains the information we need
        # otherwise skip this round
        if not check_plot_data(is_locked, plot_data):
            return

        try:
            if not is_locked:
                combined_error_signal = combine_error_signal(
                    (plot_data["error_signal_1"], plot_data.get("error_signal_2")),
                    self.parameters.dual_channel.value,
                    self.parameters.channel_mixing.value,
                    self.parameters.combined_offset.value,
                )

                if (
                    self.autolock_mode_detector is not None
                    and not self.autolock_mode_detector.done
                ):
                    self.autolock_mode_detector.handle_new_spectrum(
                        combined_error_signal
                    )
                    self.additional_spectra.append(combined_error_signal)

                    if self.autolock_mode_detector.done:
                        self.start_autolock(self.autolock_mode_detector.mode)
                    else:
                        return

                return self.algorithm.handle_new_spectrum(combined_error_signal)

            else:
                error_signal = plot_data["error_signal"]
                control_signal = plot_data["control_signal"]

                return self.after_lock(
                    error_signal, control_signal, plot_data.get("slow")
                )

        except Exception:
            traceback.print_exc()
            self.parameters.autolock_failed.value = True
            self.exposed_stop()

    def record_first_error_signal(self, error_signal, auto_offset):
        (
            mean_signal,
            target_slope_rising,
            target_zoom,
            error_signal_rolled,
            line_width,
            peak_idxs,
        ) = get_lock_point(error_signal, self.x0, self.x1)

        self.central_y = int(mean_signal)

        if auto_offset:
            self.control.pause_acquisition()
            self.parameters.combined_offset.value = -1 * self.central_y
            error_signal -= self.central_y
            error_signal_rolled -= self.central_y
            self.additional_spectra = [
                s - self.central_y for s in self.additional_spectra
            ]
            self.control.exposed_write_registers()
            self.control.continue_acquisition()

        self.parameters.target_slope_rising.value = target_slope_rising
        self.control.exposed_write_registers()

        return error_signal, error_signal_rolled, line_width, peak_idxs

    def after_lock(self, error_signal, control_signal, slow_out):
        print("after lock")
        self.parameters.autolock_locked.value = True

        self.remove_data_listener()
        self.parameters.autolock_running.value = False

        self.algorithm.after_lock()

    def relock(self):
        """
        Relock the laser using the reference spectrum recorded in the first
        locking approach.
        """
        # we check each parameter before setting it because otherwise
        # this may crash the client if called very often (e.g.if the
        # autolock continuously fails)
        if not self.parameters.autolock_running.value:
            self.parameters.autolock_running.value = True
        if not self.parameters.autolock_retrying.value:
            self.parameters.autolock_retrying.value = True

        self.reset_properties()
        self._reset_scan()

        # add a listener that listens for new spectrum data and consequently
        # tries to relock.
        self.add_data_listener()

    def exposed_stop(self):
        """Abort any operation."""
        self.parameters.autolock_preparing.value = False
        self.parameters.autolock_percentage.value = 0
        self.parameters.autolock_running.value = False
        self.parameters.autolock_locked.value = False
        self.parameters.autolock_watching.value = False
        self.parameters.fetch_additional_signals.value = True
        self.remove_data_listener()

        self._reset_scan()
        self.parameters.task.value = None

    def _reset_scan(self):
        self.control.pause_acquisition()

        self.parameters.sweep_amplitude.value = (
            self.parameters.autolock_initial_sweep_amplitude.value
        )
        self.control.exposed_start_sweep()

        self.control.continue_acquisition()



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import _thread
import os
import pickle
import sys
import threading
from random import random
from time import sleep

import click
import numpy as np
import rpyc
from linien_common.common import (
    N_POINTS,
    check_plot_data,
    pack,
    unpack,
    update_signal_history,
)
from linien_common.config import DEFAULT_SERVER_PORT
from linien_server.autolock.autolock import Autolock
from linien_server.optimization.optimization import OptimizeSpectroscopy
from linien_server.parameter_store import ParameterStore
from linien_server.parameters import Parameters
from linien_server.pid_optimization.pid_optimization import (
    PIDOptimization,
    PSDAcquisition,
)
from rpyc.utils.authenticators import AuthenticationError
from rpyc.utils.server import ThreadedServer


class BaseService(rpyc.Service):
    """A service that provides functionality for seamless integration of
    parameter access on the client."""

    def __init__(self):
        self.parameters = Parameters()
        self.parameter_store = ParameterStore(self.parameters)
        self._uuid_mapping = {}

    def on_connect(self, client):
        self._uuid_mapping[client] = client.root.uuid

    def on_disconnect(self, client):
        uuid = self._uuid_mapping[client]
        self.parameters.unregister_remote_listeners(uuid)

    def exposed_get_param(self, param_name):
        return pack(self.parameters._get_param(param_name).value)

    def exposed_set_param(self, param_name, value):
        self.parameters._get_param(param_name).value = unpack(value)

    def exposed_init_parameter_sync(self, uuid):
        return pack(list(self.parameters.init_parameter_sync(uuid)))

    def exposed_register_remote_listener(self, uuid, param_name):
        return self.parameters.register_remote_listener(uuid, param_name)

    def exposed_register_remote_listeners(self, uuid, param_names):
        for param_name in param_names:
            self.exposed_register_remote_listener(uuid, param_name)

    def exposed_get_listener_queue(self, uuid):
        return self.parameters.get_listener_queue(uuid)


class RedPitayaControlService(BaseService):
    """Control server that runs on the RP that provides high-level methods."""

    def __init__(self, **kwargs):
        self._cached_data = {}
        self.exposed_is_locked = None

        super().__init__()

        from linien_server.registers import Registers

        self.registers = Registers(**kwargs)
        self.registers.connect(self, self.parameters)

    def run_acquiry_loop(self):
        """Starts a background process that keeps polling control and error
        signal. Every received value is pushed to `parameters.to_plot`."""

        def data_received(is_raw, plot_data, data_uuid):
            # When a parameter is changed, `pause_acquisition` is set.
            # This means that the we should skip new data until we are sure that
            # it was recorded with the new settings.
            if not self.parameters.pause_acquisition.value:
                if data_uuid != self.data_uuid:
                    return

                data_loaded = pickle.loads(plot_data)

                if not is_raw:
                    is_locked = self.parameters.lock.value

                    if not check_plot_data(is_locked, data_loaded):
                        print(
                            "warning: incorrect data received for lock state, ignoring!"
                        )
                        return

                    self.parameters.to_plot.value = plot_data
                    self._generate_signal_stats(data_loaded)

                    # update signal history (if in locked state)
                    (
                        self.parameters.control_signal_history.value,
                        self.parameters.monitor_signal_history.value,
                    ) = update_signal_history(
                        self.parameters.control_signal_history.value,
                        self.parameters.monitor_signal_history.value,
                        data_loaded,
                        is_locked,
                        self.parameters.control_signal_history_length.value,
                    )
                else:
                    self.parameters.acquisition_raw_data.value = plot_data

        self.registers.run_data_acquisition(data_received)
        self.pause_acquisition()
        self.continue_acquisition()

    def run_periodic_timer(self):
        """Starts a timer that increases the `ping` parameter once per second.
        Its purpose is to allow for periodic tasks on the server:
        just register an `on_change` listener for this parameter.
        """

        def send_ping():
            while True:
                self.parameters.ping.value = self.parameters.ping.value + 1
                print("ping", self.parameters.ping.value)
                sleep(1)

        t = threading.Thread(target=send_ping)
        t.daemon = True
        t.start()
        self._periodic_timer = t

    def _generate_signal_stats(self, to_plot):
        stats = {}

        for signal_name, signal in to_plot.items():
            stats["%s_mean" % signal_name] = np.mean(signal)
            stats["%s_std" % signal_name] = np.std(signal)
            stats["%s_max" % signal_name] = np.max(signal)
            stats["%s_min" % signal_name] = np.min(signal)

        self.parameters.signal_stats.value = stats

    def exposed_write_registers(self):
        """Syncs the parameters with the FPGA registers."""
        self.registers.write_registers()

    def task_running(self):
        return (
            self.parameters.autolock_running.value
            or self.parameters.optimization_running.value
            or self.parameters.psd_acquisition_running.value
            or self.parameters.psd_optimization_running.value
        )

    def exposed_start_autolock(self, x0, x1, spectrum, additional_spectra=None):
        spectrum = pickle.loads(spectrum)
        # start_watching = self.parameters.watch_lock.value
        start_watching = False
        auto_offset = self.parameters.autolock_determine_offset.value

        if not self.task_running():
            autolock = Autolock(self, self.parameters)
            self.parameters.task.value = autolock
            autolock.run(
                x0,
                x1,
                spectrum,
                should_watch_lock=start_watching,
                auto_offset=auto_offset,
                additional_spectra=pickle.loads(additional_spectra)
                if additional_spectra is not None
                else None,
            )

    def exposed_start_optimization(self, x0, x1, spectrum):
        if not self.task_running():
            optim = OptimizeSpectroscopy(self, self.parameters)
            self.parameters.task.value = optim
            optim.run(x0, x1, spectrum)

    def exposed_start_psd_acquisition(self):
        if not self.task_running():
            self.parameters.task.value = PSDAcquisition(self, self.parameters)
            self.parameters.task.value.run()

    def exposed_start_pid_optimization(self):
        if not self.task_running():
            self.parameters.task.value = PIDOptimization(self, self.parameters)
            self.parameters.task.value.run()

    def exposed_start_sweep(self):
        self.pause_acquisition()

        self.parameters.combined_offset.value = 0
        self.parameters.lock.value = False
        self.exposed_write_registers()

        self.continue_acquisition()

    def exposed_start_lock(self):
        self.pause_acquisition()

        self.parameters.lock.value = True
        self.exposed_write_registers()

        self.continue_acquisition()

    def exposed_shutdown(self):
        """Kills the server."""
        self.registers.acquisition.shutdown()
        _thread.interrupt_main()
        # we use SystemExit instead of os._exit because we want to call atexit
        # handlers
        raise SystemExit

    def exposed_get_server_version(self):
        import linien_server

        return linien_server.__version__

    def exposed_get_restorable_parameters(self):
        return self.parameters._restorable_parameters

    def exposed_pause_acquisition(self):
        self.pause_acquisition()

    def exposed_continue_acquisition(self):
        self.continue_acquisition()

    def exposed_set_csr_direct(self, k, v):
        """Directly sets a CSR register. This method is intended for debugging.
        Normally, the FPGA should be controlled via manipulation of parameters."""
        self.registers.set(k, v)

    def pause_acquisition(self):
        """Pause continuous acquisition. Call this before changing a parameter
        that alters the error / control signal. This way, no inconsistent signals
        reach the application. After setting the new parameter values, call
        `continue_acquisition`."""
        self.parameters.pause_acquisition.value = True
        self.data_uuid = random()
        self.registers.acquisition.pause_acquisition()

    def continue_acquisition(self):
        """Continue acquisition after a short delay, when we are sure that the
        new parameters values have been written to the FPGA and that data that
        is now recorded is recorded with the correct parameters."""
        self.parameters.pause_acquisition.value = False
        self.registers.acquisition.continue_acquisition(self.data_uuid)


class FakeRedPitayaControl(BaseService):
    def __init__(self):
        super().__init__()
        self.exposed_is_locked = None

    def exposed_write_registers(self):
        pass

    def run_acquiry_loop(self):
        import threading
        from random import randint
        from time import sleep

        def run():
            while True:
                max_ = randint(0, 8191)
                gen = lambda: np.array([randint(-max_, max_) for _ in range(N_POINTS)])
                self.parameters.to_plot.value = pickle.dumps(
                    {
                        "error_signal_1": gen(),
                        "error_signal_1_quadrature": gen(),
                        "error_signal_2": gen(),
                        "error_signal_2_quadrature": gen(),
                    }
                )
                sleep(0.1)

        t = threading.Thread(target=run)
        t.daemon = True
        t.start()

    def run_periodic_timer(self):
        pass

    def exposed_shutdown(self):
        _thread.interrupt_main()
        os._exit(0)

    def exposed_start_autolock(self, x0, x1, spectrum):
        print("start autolock", x0, x1)

    def exposed_start_optimization(self, x0, x1, spectrum):
        print("start optimization")
        self.parameters.optimization_running.value = True

    def exposed_get_restorable_parameters(self):
        return self.parameters._restorable_parameters

    def exposed_get_server_version(self):
        import linien_server

        return linien_server.__version__

    def pause_acquisition(self):
        pass

    def continue_acquisition(self):
        pass


@click.command()
@click.argument("port", default=DEFAULT_SERVER_PORT, type=int, required=False)
@click.option(
    "--fake", is_flag=True, help="Runs a fake server that just returns random data"
)
@click.option(
    "--remote-rp",
    help="Allows to run the server locally for development and "
    "connects to a RedPitaya. Specify the RP's credentials "
    "as follows: "
    "--remote-rp=root:myPassword@rp-f0xxxx.local",
)
def run_server(port, fake=False, remote_rp=False):
    print("start server at port", port)

    if fake:
        print("starting fake server")
        control = FakeRedPitayaControl()
    else:
        if remote_rp is not None:
            assert (
                "@" in remote_rp and ":" in remote_rp
            ), "invalid format, should be root:myPassword@rp-f0xxxx.local"

            username, tmp = remote_rp.split(":", 1)
            r_host, r_password = "".join(reversed(tmp)).split("@", 1)
            host = "".join(reversed(r_host))
            password = "".join(reversed(r_password))
            control = RedPitayaControlService(
                host=host, user=username, password=password
            )
        else:
            control = RedPitayaControlService()

    control.run_acquiry_loop()
    control.run_periodic_timer()
    control.exposed_write_registers()

    failed_auth_counter = {"c": 0}

    def username_and_password_authenticator(sock):
        # when a client starts the server, it supplies this hash via an environment
        # variable
        secret = os.environ.get("LINIEN_AUTH_HASH")

        # client always sends auth hash, even if we run in non-auth mode
        # --> always read 64 bytes, otherwise rpyc connection can't be established
        received = sock.recv(64)

        # as a protection against brute force, we don't accept requests after
        # too many failed auth requests
        if failed_auth_counter["c"] > 1000:
            print("received too many failed auth requests!")
            sys.exit(1)

        if secret is None:
            print("warning: no authentication set up")
        else:
            if received != secret.encode():
                print("received invalid credentials: ", received)

                failed_auth_counter["c"] += 1

                raise AuthenticationError("invalid username / password")

            print("authentication successful")

        return sock, None

    t = ThreadedServer(
        control,
        port=port,
        authenticator=username_and_password_authenticator,
        protocol_config={"allow_pickle": True},
    )
    t.start()


if __name__ == "__main__":
    run_server()



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from linien_common.common import (
    ANALOG_OUT0,
    HIGH_PASS_FILTER,
    LOW_PASS_FILTER,
    MHz,
    convert_channel_mixing_value,
)
from linien_common.config import DEFAULT_SWEEP_SPEED

from .acquisition import AcquisitionMaster
from .csr import PitayaCSR
from .iir_coeffs import make_filter
from .utils import twos_complement


class Registers:
    """This class provides low-level access to the FPGA registers.

    High-level applications should not access this class directly but instead
    communicate by manipulating `Parameters` / `RemoteParameters`.
    """

    def __init__(self, host=None, user=None, password=None):
        self.host = host
        self.user = user
        self.password = password
        self.acquisition = None

        self._last_sweep_speed = None
        self._last_raw_acquisition_settings = None
        self._iir_cache = {}

    def connect(self, control, parameters):
        """Starts a process that can be used to control FPGA registers."""
        self.control = control
        self.parameters = parameters

        self.csr = PitayaCSR()

        def lock_status_changed(v):
            if self.acquisition is not None:
                self.acquisition.lock_status_changed(v)

        self.parameters.lock.on_change(lock_status_changed)

        def fetch_additional_signals_changed(v):
            if self.acquisition is not None:
                self.acquisition.fetch_additional_signals_changed(v)

        self.parameters.fetch_additional_signals.on_change(
            fetch_additional_signals_changed
        )

        def dual_channel_changed(dual_channel):
            if self.acquisition is not None:
                self.acquisition.set_dual_channel(dual_channel)

        self.parameters.dual_channel.on_change(dual_channel_changed)

        use_ssh = self.host is not None and self.host not in ("localhost", "127.0.0.1")
        self.acquisition = AcquisitionMaster(use_ssh, self.host)

    def run_data_acquisition(self, on_change):
        """Starts a background process that continuously reads out error /
        control signal of the FPGA. For every result, `on_change` is called."""
        self.acquisition.run_data_acquisition(on_change)

    def write_registers(self):
        """Writes data from `parameters` to the FPGA."""
        params = dict(self.parameters)

        _max = lambda val: val if np.abs(val) <= 8191 else (8191 * val / np.abs(val))
        phase_to_delay = lambda phase: int(phase / 360 * (1 << 14))

        if not params["dual_channel"]:
            factor_a = 256
            factor_b = 0
        else:
            value = params["channel_mixing"]
            factor_a, factor_b = convert_channel_mixing_value(value)

        lock = params["lock"]
        lock_changed = lock != self.control.exposed_is_locked
        self.control.exposed_is_locked = lock

        new = dict(
            # sweep run is 1 by default. The gateware automatically takes care
            # of stopping the sweep run after `request_lock` is set by setting
            # `sweep.clear`
            logic_sweep_run=1,
            logic_sweep_pause=int(params["sweep_pause"]),
            logic_sweep_step=int(
                DEFAULT_SWEEP_SPEED
                * params["sweep_amplitude"]
                / (2 ** params["sweep_speed"])
            ),
            # NOTE: Sweep center is set by `logic_out_offset`.
            logic_sweep_min=-1 * _max(params["sweep_amplitude"] * 8191),
            logic_sweep_max=_max(params["sweep_amplitude"] * 8191),
            logic_mod_freq=params["modulation_frequency"],
            logic_mod_amp=params["modulation_amplitude"]
            if params["modulation_frequency"] > 0
            else 0,
            logic_dual_channel=int(params["dual_channel"]),
            logic_fast_mode=int(params["fast_mode"]),
            logic_chain_a_factor=factor_a,
            logic_chain_b_factor=factor_b,
            logic_chain_a_offset=twos_complement(int(params["offset_a"]), 14),
            logic_chain_b_offset=twos_complement(int(params["offset_b"]), 14),
            logic_out_offset=int(params["sweep_center"] * 8191),
            logic_combined_offset=twos_complement(params["combined_offset"], 14),
            logic_control_channel=params["control_channel"],
            logic_mod_channel=params["mod_channel"],
            logic_sweep_channel=params["sweep_channel"],
            slow_pid_reset=not params["pid_on_slow_enabled"],
            logic_analog_out_1=params["analog_out_1"],
            logic_analog_out_2=params["analog_out_2"],
            logic_analog_out_3=params["analog_out_3"],
            logic_autolock_fast_target_position=params["autolock_target_position"],
            logic_autolock_autolock_mode=params["autolock_mode"],
            logic_autolock_robust_N_instructions=len(params["autolock_instructions"]),
            logic_autolock_robust_time_scale=params["autolock_time_scale"],
            logic_autolock_robust_final_wait_time=params["autolock_final_wait_time"],
            # channel A
            fast_a_demod_delay=phase_to_delay(params["demodulation_phase_a"])
            if params["modulation_frequency"] > 0
            else 0,
            fast_a_demod_multiplier=params["demodulation_multiplier_a"],
            fast_a_dx_sel=self.csr.signal("zero"),
            fast_a_y_tap=2,
            fast_a_dy_sel=self.csr.signal("zero"),
            fast_a_invert=int(params["invert_a"]),
            # channel B
            fast_b_demod_delay=phase_to_delay(params["demodulation_phase_b"])
            if params["modulation_frequency"] > 0
            else 0,
            fast_b_demod_multiplier=params["demodulation_multiplier_b"],
            fast_b_dx_sel=self.csr.signal("zero"),
            fast_b_y_tap=1,
            fast_b_dy_sel=self.csr.signal("zero"),
            fast_b_invert=int(params["invert_b"]),
            # trigger on sweep
            scopegen_external_trigger=1,
            gpio_p_oes=0b11111111,
            gpio_n_oes=0b11111111,
            gpio_p_outs=params["gpio_p_out"],
            gpio_n_outs=params["gpio_n_out"],
            gpio_n_do0_en=self.csr.signal("zero"),
            gpio_n_do1_en=self.csr.signal("zero"),
            logic_slow_decimation=16,
        )

        for instruction_idx, [wait_for, peak_height] in enumerate(
            params["autolock_instructions"]
        ):
            new["logic_autolock_robust_peak_height_%d" % instruction_idx] = peak_height
            new["logic_autolock_robust_wait_for_%d" % instruction_idx] = wait_for

        if lock:
            # display combined error signal and control signal
            new.update(
                {
                    "scopegen_adc_a_sel": self.csr.signal(
                        "logic_combined_error_signal"
                        if not params["acquisition_raw_filter_enabled"]
                        else "logic_combined_error_signal_filtered"
                    ),
                    "scopegen_adc_a_q_sel": self.csr.signal("fast_b_x"),
                    "scopegen_adc_b_sel": self.csr.signal("logic_control_signal"),
                    "scopegen_adc_b_q_sel": self.csr.signal("zero"),
                }
            )
        else:
            # display both demodulated error signals (if dual channel mode)
            # OR: display demodulated error signal 1 + monitor signal
            new.update(
                {
                    "scopegen_adc_a_sel": self.csr.signal("fast_a_out_i"),
                    "scopegen_adc_a_q_sel": self.csr.signal("fast_a_out_q"),
                    "scopegen_adc_b_sel": self.csr.signal(
                        "fast_b_out_i" if params["dual_channel"] else "fast_b_x"
                    ),
                    "scopegen_adc_b_q_sel": self.csr.signal(
                        "fast_b_out_q" if params["dual_channel"] else "zero"
                    ),
                }
            )

        # filter out values that did not change
        new = dict(
            (k, v)
            for k, v in new.items()
            if (
                (k not in self.control._cached_data)
                or (self.control._cached_data.get(k) != v)
            )
        )
        self.control._cached_data.update(new)

        # pass sweep speed changes to acquisition process
        sweep_changed = params["sweep_speed"] != self._last_sweep_speed
        if sweep_changed:
            self._last_sweep_speed = params["sweep_speed"]
            self.acquisition.set_sweep_speed(params["sweep_speed"])

        raw_acquisition_settings = (
            params["acquisition_raw_enabled"],
            params["acquisition_raw_decimation"],
        )
        if raw_acquisition_settings != self._last_raw_acquisition_settings:
            self._last_raw_acquisition_settings = raw_acquisition_settings
            self.acquisition.set_raw_acquisition(*raw_acquisition_settings)

        fpga_base_freq = 125e6

        self.set_iir(
            "logic_raw_acquisition_iir",
            *make_filter(
                "LP", f=params["acquisition_raw_filter_frequency"] / fpga_base_freq, k=1
            )
        )

        for k, v in new.items():
            self.set(k, int(v))

        if not lock and sweep_changed:
            # reset sweep for a short time if the scan range was changed
            # this is needed because otherwise it may take too long before
            # the new scan range is reached --> no scope trigger is sent
            self.set("logic_sweep_run", 0)
            self.set("logic_sweep_run", 1)

        kp = params["p"]
        ki = params["i"]
        kd = params["d"]
        slope = params["target_slope_rising"]
        control_channel, sweep_channel = (
            params["control_channel"],
            params["sweep_channel"],
        )

        def channel_polarity(channel):
            return (
                params["polarity_fast_out1"],
                params["polarity_fast_out2"],
                params["polarity_analog_out0"],
            )[channel]

        if control_channel != sweep_channel:
            if channel_polarity(control_channel) != channel_polarity(sweep_channel):
                slope = not slope

        slow_strength = (
            params["pid_on_slow_strength"] if params["pid_on_slow_enabled"] else 0
        )
        slow_slope = (
            1
            if channel_polarity(ANALOG_OUT0) == channel_polarity(control_channel)
            else -1
        )

        for chain in ("a", "b"):
            automatic = params["filter_automatic_%s" % chain]
            # iir_idx means iir_c or iir_d
            for iir_idx in range(2):
                # iir_sub_idx means in-phase signal or quadrature signal
                for iir_sub_idx in range(2):
                    iir_name = "fast_%s_iir_%s_%d" % (
                        chain,
                        ("c", "d")[iir_idx],
                        iir_sub_idx + 1,
                    )

                    if automatic:
                        filter_enabled = True
                        filter_type = LOW_PASS_FILTER
                        filter_frequency = (
                            params["modulation_frequency"] / MHz * 1e6 / 2
                        )

                        # if the filter frequency is too low (< 10Hz), the IIR doesn't
                        # work properly anymore. In that case, don't filter.
                        # This is also helpful if the raw (not demodulated) signal
                        # should be displayed which can be achieved by setting
                        # modulation frequency to 0.
                        if filter_frequency < 10:
                            filter_enabled = False
                    else:
                        filter_enabled = params[
                            "filter_%d_enabled_%s" % (iir_idx + 1, chain)
                        ]
                        filter_type = params["filter_%d_type_%s" % (iir_idx + 1, chain)]
                        filter_frequency = params[
                            "filter_%d_frequency_%s" % (iir_idx + 1, chain)
                        ]

                    if not filter_enabled:
                        self.set_iir(iir_name, *make_filter("P", k=1))
                    else:
                        if filter_type == LOW_PASS_FILTER:
                            self.set_iir(
                                iir_name,
                                *make_filter(
                                    "LP", f=filter_frequency / fpga_base_freq, k=1
                                )
                            )
                        elif filter_type == HIGH_PASS_FILTER:
                            self.set_iir(
                                iir_name,
                                *make_filter(
                                    "HP", f=filter_frequency / fpga_base_freq, k=1
                                )
                            )
                        else:
                            raise Exception(
                                "unknown filter %s for %s" % (filter_type, iir_name)
                            )

        if lock_changed:
            if lock:
                # set PI parameters
                self.set_pid(kp, ki, kd, slope, reset=0, request_lock=1)
                self.set_slow_pid(slow_strength, slow_slope, reset=0)
            else:
                self.set_pid(0, 0, 0, slope, reset=1, request_lock=0)
                self.set_slow_pid(0, slow_slope, reset=1)
        else:
            if lock:
                # set new PI parameters
                self.set_pid(kp, ki, kd, slope)
                self.set_slow_pid(slow_strength, slow_slope)

    def set_pid(self, p, i, d, slope, reset=None, request_lock=None):
        if request_lock is not None:
            self.set("logic_autolock_request_lock", request_lock)

        sign = -1 if slope else 1
        self.set("logic_pid_kp", p * sign)
        self.set("logic_pid_ki", i * sign)
        self.set("logic_pid_kd", d * sign)

        if reset is not None:
            self.set("logic_pid_reset", reset)

    def set_slow_pid(self, strength, slope, reset=None):
        sign = slope
        self.set("slow_pid_ki", strength * sign)

        if reset is not None:
            self.set("slow_pid_reset", reset)

    def set(self, key, value):
        self.acquisition.set_csr(key, value)

    def set_iir(self, iir_name, *args):
        if self._iir_cache.get(iir_name) != args:
            # as setting iir parameters takes some time, take care that we don't
            # do it too often
            self.acquisition.set_iir_csr(iir_name, *args)
            self._iir_cache[iir_name] = args



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

from linien_common.common import (
    AUTO_DETECT_AUTOLOCK_MODE,
    FAST_AUTOLOCK,
    PSD_ALGORITHM_LPSD,
    MHz,
    Vpp,
)
from linien_common.config import DEFAULT_COLORS, N_COLORS
from linien_server.parameters_base import BaseParameters, Parameter


class Parameters(BaseParameters):
    """This class defines the parameters of the Linien server. They represent
    the public interface and can be used to control the behavior of the server.

    Access on the server takes place like this:

        # retrieve a parameter
        foo(parameters.modulation_amplitude.value)

        # set a parameter
        parameters.modulation_amplitude.value = 0.5 * Vpp

        # if a parameter influences the behavior of the FPGA, it has to be
        # written to the FPGA as well (`control` is an instance of
        # `RedPitayaControlService`):
        control.exposed_write_registers()

    On the client side, access happens through `RemoteParameters` which
    transparently mimics the behavior of this class. Have a look at the comments
    below for a description of each parameter.
    """

    def __init__(self):
        super().__init__()

        # parameters whose values are saved by the client and restored if the
        # client connects to the RedPitaya with no server running.
        self._restorable_parameters = (
            "fast_mode",
            "modulation_amplitude",
            "modulation_frequency",
            "sweep_speed",
            "demodulation_phase_a",
            "demodulation_multiplier_a",
            "demodulation_phase_b",
            "demodulation_multiplier_b",
            "offset_a",
            "offset_b",
            "invert_a",
            "invert_b",
            "filter_1_enabled_a",
            "filter_1_enabled_b",
            "filter_1_frequency_a",
            "filter_1_frequency_b",
            "filter_1_type_a",
            "filter_1_type_b",
            "filter_2_enabled_a",
            "filter_2_enabled_b",
            "filter_2_frequency_a",
            "filter_2_frequency_b",
            "filter_2_type_a",
            "filter_2_type_b",
            "filter_automatic_a",
            "filter_automatic_b",
            "p",
            "i",
            "d",
            "watch_lock",
            "watch_lock_threshold",
            "dual_channel",
            "channel_mixing",
            "pid_on_slow_enabled",
            "pid_on_slow_strength",
            "mod_channel",
            "control_channel",
            "sweep_channel",
            "polarity_fast_out1",
            "polarity_fast_out2",
            "polarity_analog_out0",
            "check_lock",
            "analog_out_1",
            "analog_out_2",
            "analog_out_3",
            "plot_line_width",
            "plot_color_0",
            "plot_color_1",
            "plot_color_2",
            "plot_color_3",
            "plot_line_opacity",
            "plot_fill_opacity",
            "autolock_determine_offset",
            "autolock_mode_preference",
            "psd_acquisition_max_decimation",
        )

        # the `to_plot` parameter is a pickled dictionary that contains signals
        # that may be plotted. Depending on the locking state, it may contain these
        # signals:
        # Unlocked state:
        #   - `error_signal_1` and `error_signal_1_quadrature`:
        #       IQ-demodulated and low-pass-filtered error signals from
        #       ANALOG IN 0
        #   - `error_signal_2` and `error_signal_2_quadrature`:
        #       IQ-demodulated and low-pass-filtered error signals from
        #       ANALOG IN 1. These signals are only available if in dual-channel
        #       spectroscopy mode is enabled. Otherwise the un-demodulated
        #       monitor_signal` is shown.
        #   - `monitor_signal`:
        #       Signal recorded from ANALOG IN 1 without demodulation. Only
        #       available if dual channel mode is not enabled.
        #
        # Locked state:
        #   - `error_signal`:
        #       Error signal that is fed into the PID controller.
        #   - `control_signal`:
        #       Output of the PID controller.
        #   - `slow`:
        #       Output of slow additional integrator on slow analog output.
        #       Note that this value is not an array but a single point.
        self.to_plot = Parameter(sync=False)
        # A dictionary that contains mean, standard deviation, minimum value and
        # maximum value for all the signals contained in `to_plot`.
        # Exemplary dictionary keys are `error_signal_mean`,
        # `control_signal_std`, `monitor_signal_min` or `error_signal_2_max`.
        self.signal_stats = Parameter(sync=False)

        #           --------- GENERAL PARAMETERS ---------
        # configures the output of the modulation frequency. A value of 0 means
        # FAST OUT 1 and a value of 1 corresponds to FAST OUT 2
        self.mod_channel = Parameter(start=0, min_=0, max_=1)
        # configures the output of the scan sweep:
        #       0 --> FAST OUT 1
        #       1 --> FAST OUT 2
        #       2 --> ANALOG OUT 0 (slow channel)
        self.sweep_channel = Parameter(start=1, min_=0, max_=2)
        # configures the output of the lock signal. A value of 0 means
        # FAST OUT 1 and a value of 1 corresponds to FAST OUT 2
        self.control_channel = Parameter(start=1, min_=0, max_=1)

        # set the output of GPIO pins. Each bit corresponds to one pin, i.e.:
        #       parameters.gpio_p_out.value = 0b11110000
        # turns on the first 4 pins and turns off the other ones.
        self.gpio_p_out = Parameter(start=0, min_=0, max_=0b11111111)
        self.gpio_n_out = Parameter(start=0, min_=0, max_=0b11111111)

        # parameters for setting ANALOG_OUT voltage
        # usage:
        #       parameters.analog_out_1.value = 1.2 * ANALOG_OUT_V
        # Minimum value is 0 and maximum 1.8 * ANALOG_OUT_V
        # note that ANALOG_OUT_0 ís used for the slow PID and thus can't be
        for i in range(4):
            if i == 0:
                # ANALOG_OUT0 is used for slow PID --> it can't be controlled
                # manually
                continue

            setattr(
                self,
                "analog_out_%d" % i,
                Parameter(start=0, min_=0, max_=(2**15) - 1),
            )

        # If `True`, this parameter turns off the sweep and starts the PID
        self.lock = Parameter(start=False)

        # for both fast outputs and the analog out, define whether tuning the
        # voltage up correspond to tuning the laser frequency up or down. Setting
        # these values correctly is only required when using both, a fast out and
        # a the slow analog output for PID
        self.polarity_fast_out1 = Parameter(start=False)
        self.polarity_fast_out2 = Parameter(start=False)
        self.polarity_analog_out0 = Parameter(start=False)

        # record of control signal should be kept for how long?
        self.control_signal_history_length = Parameter(start=600)
        self.control_signal_history = Parameter(
            start={"times": [], "values": []}, sync=False
        )
        self.monitor_signal_history = Parameter(
            start={"times": [], "values": []}, sync=False
        )
        # if this boolean is `True`, no new spectroscopy data is sent to the
        # clients. This parameter is used when writing data to FPGA that would
        # result in cropped / distorted signals being displayed.
        self.pause_acquisition = Parameter(start=False)

        # this parameter is not exposed to GUI. It is used by the autolock or
        # normal lock to fetch less data if they are not needed.
        self.fetch_additional_signals = Parameter(start=True)

        # this is just a counter that is automatically increased every second.
        # Its purpose is to allow for periodic tasks on the server:
        # just register an `on_change` listener for this parameter.
        self.ping = Parameter(start=0)

        #           --------- SWEEP PARAMETERS ---------

        # Amplitude of the sweep in units of 0.5 * Vpp of the output (2 V for fast
        # outputs (range +/- 1 V) and 0.9 V for slow outputs (range 0 V to 1.8 V). That
        # means an amplitude of 1.0 corresponds to the full sweep range in both cases.
        self.sweep_amplitude = Parameter(min_=0.001, max_=1, start=1)
        # The center position of the sweep. If a fast output is used for the sweep this
        #  is the sweep center position in volts. If the slow output is used the
        # interval [-1, +1] of this parameter is mapped to the interval [0V, +1.8V].
        self.sweep_center = Parameter(min_=-1, max_=1, start=0)
        # The sweep speed in internal units. The actual speed is given by
        #       f_real = 3.8 kHz / (2 ** sweep_speed)
        # Allowed values are [0, ..., 16]
        self.sweep_speed = Parameter(min_=0, max_=32, start=8)
        # If set to `True`, this parameter pauses the sweep at the `sweep_center`.
        self.sweep_pause = Parameter(start=False)

        #           --------- MODULATION PARAMETERS ---------

        # The amplitude of the modulation in internal units. Use Vpp for
        # conversion to volts peak-peak, e.g:
        #       parameters.modulation_amplitude.value = 0.5 * Vpp
        # Values between 0 and 2 * Vpp are allowed.
        self.modulation_amplitude = Parameter(min_=0, max_=(1 << 14) - 1, start=1 * Vpp)

        # Frequency of the modulation in internal units. Use MHz for conversion
        # to human-readable frequency, e.g:
        #       parameters.modulation_frequency.value = 6.6 * MHz
        # By design, values up to 128 * MHz = 0xffffffff are allowed although in
        # practice values of more than 50 MHz don't make sense due to the limited
        # sampling rate of the DAC.
        self.modulation_frequency = Parameter(min_=0, max_=0xFFFFFFFF, start=15 * MHz)

        #           --------- DEMODULATION AND FILTER PARAMETERS ---------
        # fast mode allows to bypass demodulation and IIR filtering of the fast
        # channels.
        self.fast_mode = Parameter(start=False)
        # Linien allows for two simulataneous demodulation channels. By default,
        # only one is enabled. This is controlled by `dual_channel`.
        self.dual_channel = Parameter(start=False)
        # If in dual channel mode, what is the mixing ratio between them?
        # A value of 0 corresponds to equal ratio
        #            -128             only channel A being active
        #            128              only channel B being active
        # Integer values [-128, ..., 128] are allowed.
        self.channel_mixing = Parameter(start=0)

        # The following parameters exist twice, i.e. once per channel
        for channel in ("a", "b"):
            # The demodulation phase in degree (0-360)
            setattr(
                self,
                "demodulation_phase_%s" % channel,
                Parameter(min_=0, max_=360, start=0x0, wrap=True),
            )
            # This parameter allows for multi-f (e.g. 3f or 5f) demodulation.
            # Default value is 1, indicating that 1f demodulation is used.
            setattr(
                self,
                "demodulation_multiplier_%s" % channel,
                Parameter(min_=0, max_=15, start=1),
            )
            # The vertical offset for a channel. A value of -8191 shifts the data
            # down by 1V, a value of +8191 moves it up.
            setattr(
                self, "offset_%s" % channel, Parameter(min_=-8191, max_=8191, start=0)
            )
            # A boolean indicating whether the channel data should be inverted.
            setattr(self, "invert_%s" % channel, Parameter(start=False))

            # - -----   FILTER PARAMETERS   -----
            # after demodulation of the signal, Linien may apply up to two IIR
            # filters.
            # `filter_automatic` is a boolean indicating whether Linien should
            # automatically determine suitable filter for a given modulation
            # frequency or whether the user may configure the filters himself.
            # If automatic mode is enabled, two low pass filters are installed
            # with a frequency of half the modulation frequency.
            setattr(self, "filter_automatic_%s" % channel, Parameter(start=True))

            for filter_i in [1, 2]:
                # should this filter be enabled? Note that disabling a filter
                # does not bypass it as this would change the propagation time
                # of the signal through the FPGA which is unfavorable as it leads
                # to a mismatch of the demodulation phase. Instead, a filter
                # with unity gain is installed.
                setattr(
                    self,
                    "filter_%d_enabled_%s" % (filter_i, channel),
                    Parameter(start=False),
                )
                # Either `LOW_PASS_FILTER` or `HIGH_PASS_FILTER` from
                # `linien_common.common` module.
                setattr(
                    self, "filter_%d_type_%s" % (filter_i, channel), Parameter(start=0)
                )
                # The filter frequency in units of Hz
                setattr(
                    self,
                    "filter_%d_frequency_%s" % (filter_i, channel),
                    Parameter(start=10000),
                )

        #           --------- LOCK AND PID PARAMETERS ---------
        # after combining channels A and B and before passing the result to the
        # PID, `combined_offset` is added. It uses the same units as the channel
        # offsets, i.e. a value of -8191 shifts the data down by 1V, a value
        # of +8191 moves it up.
        self.combined_offset = Parameter(min_=-8191, max_=8191, start=0)
        # PID parameters. Range is [0, 8191]. In order to change sign of PID
        # parameters, use `target_slope_rising`
        self.p = Parameter(start=50, max_=8191)
        self.i = Parameter(start=5, max_=8191)
        self.d = Parameter(start=0, max_=8191)
        # A boolean that inverts the sign of the PID parameters
        self.target_slope_rising = Parameter(start=True)

        # Whether the PID on ANALOG_OUT 0 is enabled
        self.pid_on_slow_enabled = Parameter(start=False)
        # Strength of the (slow) PID on ANALOG_OUT 0. This strength corresponds
        # to the strength of the integrator. Maximum value is 8191.
        self.pid_on_slow_strength = Parameter(start=0)

        self.check_lock = Parameter(start=True)
        self.watch_lock = Parameter(start=True)
        self.watch_lock_threshold = Parameter(start=0.01)

        #           --------- AUTOLOCK PARAMETERS ---------
        # these are used internally by the autolock and usually should not be
        # manipulated
        self.task = Parameter(start=None, sync=False)
        self.automatic_mode = Parameter(start=True)
        self.autolock_target_position = Parameter(start=0)
        self.autolock_mode_preference = Parameter(start=AUTO_DETECT_AUTOLOCK_MODE)
        self.autolock_mode = Parameter(start=FAST_AUTOLOCK)
        self.autolock_time_scale = Parameter(start=0)
        self.autolock_instructions = Parameter(start=[], sync=False)
        self.autolock_final_wait_time = Parameter(start=0)
        self.autolock_selection = Parameter(start=False)
        self.autolock_running = Parameter(start=False)
        self.autolock_preparing = Parameter(start=False)
        self.autolock_percentage = Parameter(start=0, min_=0, max_=100)
        self.autolock_watching = Parameter(start=False)
        self.autolock_failed = Parameter(start=False)
        self.autolock_locked = Parameter(start=False)
        self.autolock_retrying = Parameter(start=False)
        self.autolock_determine_offset = Parameter(start=True)
        self.autolock_initial_sweep_amplitude = Parameter(start=1)

        #           --------- OPTIMIZATION PARAMETERS ---------
        # these are used internally by the optimization algorithm and usually
        # should not be manipulated
        self.optimization_selection = Parameter(start=False)
        self.optimization_running = Parameter(start=False)
        self.optimization_approaching = Parameter(start=False)
        self.optimization_improvement = Parameter(start=0)
        self.optimization_mod_freq_enabled = Parameter(start=1)
        self.optimization_mod_freq_min = Parameter(start=0.0)
        self.optimization_mod_freq_max = Parameter(start=10.0)
        self.optimization_mod_amp_enabled = Parameter(start=1)
        self.optimization_mod_amp_min = Parameter(start=0.0)
        self.optimization_mod_amp_max = Parameter(start=2.0)
        self.optimization_optimized_parameters = Parameter(start=(0, 0, 0))
        self.optimization_channel = Parameter(start=0)
        self.optimization_failed = Parameter(start=False)

        #           --------- PID OPTIMIZATION PARAMETERS ---------
        # these are used internally by the PID optimization algorithm and usually
        # should not be manipulated
        self.acquisition_raw_enabled = Parameter(start=False)
        self.acquisition_raw_decimation = Parameter(start=1)
        self.acquisition_raw_data = Parameter()
        # raw acquisition has an additiona iir filter that can be used as low
        # pass for preventing alias effects
        self.acquisition_raw_filter_enabled = Parameter(start=False)
        # The filter frequency in units of Hz
        self.acquisition_raw_filter_frequency = Parameter(start=0)
        self.psd_data_partial = Parameter(start=None)
        self.psd_data_complete = Parameter(start=None)
        self.psd_algorithm = Parameter(start=PSD_ALGORITHM_LPSD)
        self.psd_acquisition_running = Parameter(start=False)
        self.psd_optimization_running = Parameter(start=False)
        self.psd_acquisition_max_decimation = Parameter(start=18, min_=1, max_=32)

        #           --------- PARAMETERS OF GUI ---------
        self.plot_line_width = Parameter(start=2, min_=0.1, max_=100)
        self.plot_line_opacity = Parameter(start=230, min_=0, max_=255)
        self.plot_fill_opacity = Parameter(start=70, min_=0, max_=255)

        for color_idx in range(N_COLORS):
            setattr(
                self,
                "plot_color_%d" % color_idx,
                Parameter(start=DEFAULT_COLORS[color_idx]),
            )



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import atexit
import sys
import threading
from enum import Enum
from multiprocessing import Pipe, Process
from time import sleep

import rpyc
from linien_common.config import ACQUISITION_PORT
from linien_server.utils import flash_fpga, start_nginx, stop_nginx


class AcquisitionConnectionError(Exception):
    pass


class AcquisitionProcessSignals(Enum):
    SHUTDOWN = 0
    SET_SWEEP_SPEED = 2
    SET_LOCK_STATUS = 3
    SET_CSR = 4
    SET_IIR_CSR = 5
    PAUSE_ACQUISIITON = 5.5
    CONTINUE_ACQUISITION = 6
    FETCH_QUADRATURES = 7
    SET_RAW_ACQUISITION = 8
    SET_DUAL_CHANNEL = 9


class AcquisitionMaster:
    def __init__(self, use_ssh, host):
        self.on_acquisition = None

        def receive_acquired_data(conn):
            while True:
                is_raw, received_data, data_uuid = conn.recv()
                if self.on_acquisition is not None:
                    self.on_acquisition(is_raw, received_data, data_uuid)

        self.acq_process, child_pipe = Pipe()
        p = Process(
            target=self.connect_acquisition_process, args=(child_pipe, use_ssh, host)
        )
        p.daemon = True
        p.start()

        # wait until connection is established
        self.acq_process.recv()

        t = threading.Thread(target=receive_acquired_data, args=(self.acq_process,))
        t.daemon = True
        t.start()

        atexit.register(self.shutdown)

    def run_data_acquisition(self, on_acquisition):
        self.on_acquisition = on_acquisition

    def connect_acquisition_process(self, pipe, use_ssh, host):
        if use_ssh:
            # for debugging, acquisition process may be launched manually on the
            # server and rpyc can be used to connect to it
            acquisition_rpyc = rpyc.connect(host, ACQUISITION_PORT)
            acquisition = acquisition_rpyc.root
        else:
            # this is what happens in production mode
            from linien_server.acquisition_process import DataAcquisitionService

            stop_nginx()
            flash_fpga()
            acquisition = DataAcquisitionService()

        # tell the main thread that we're ready
        pipe.send(True)

        # run a loop that listens for acquired data and transmits them
        # to the main thread. Also redirects calls from the main thread
        # to the acquiry process.
        last_hash = None
        while True:
            # check whether the main thread sent a command to the acquiry process
            while pipe.poll():
                data = pipe.recv()
                if data[0] == AcquisitionProcessSignals.SHUTDOWN:
                    raise SystemExit()
                elif data[0] == AcquisitionProcessSignals.SET_SWEEP_SPEED:
                    speed = data[1]
                    acquisition.exposed_set_sweep_speed(speed)
                elif data[0] == AcquisitionProcessSignals.SET_LOCK_STATUS:
                    acquisition.exposed_set_lock_status(data[1])
                elif data[0] == AcquisitionProcessSignals.FETCH_QUADRATURES:
                    acquisition.exposed_set_fetch_additional_signals(data[1])
                elif data[0] == AcquisitionProcessSignals.SET_RAW_ACQUISITION:
                    acquisition.exposed_set_raw_acquisition(data[1])
                elif data[0] == AcquisitionProcessSignals.SET_DUAL_CHANNEL:
                    acquisition.exposed_set_dual_channel(data[1])
                elif data[0] == AcquisitionProcessSignals.SET_CSR:
                    acquisition.exposed_set_csr(*data[1])
                elif data[0] == AcquisitionProcessSignals.SET_IIR_CSR:
                    acquisition.exposed_set_iir_csr(*data[1])
                elif data[0] == AcquisitionProcessSignals.PAUSE_ACQUISIITON:
                    acquisition.exposed_pause_acquisition()
                elif data[0] == AcquisitionProcessSignals.CONTINUE_ACQUISITION:
                    acquisition.exposed_continue_acquisition(data[1])

            # load acquired data and send it to the main thread
            (
                new_data_returned,
                new_hash,
                data_was_raw,
                new_data,
                data_uuid,
            ) = acquisition.exposed_return_data(last_hash)
            if new_data_returned:
                last_hash = new_hash
                pipe.send((data_was_raw, new_data, data_uuid))

            sleep(0.05)

    def shutdown(self):
        if self.acq_process:
            self.acq_process.send((AcquisitionProcessSignals.SHUTDOWN,))

        start_nginx()

    def set_sweep_speed(self, speed):
        self.acq_process.send((AcquisitionProcessSignals.SET_SWEEP_SPEED, speed))

    def lock_status_changed(self, status):
        if self.acq_process:
            self.acq_process.send((AcquisitionProcessSignals.SET_LOCK_STATUS, status))

    def fetch_additional_signals_changed(self, status):
        if self.acq_process:
            self.acq_process.send((AcquisitionProcessSignals.FETCH_QUADRATURES, status))

    def set_csr(self, key, value):
        self.acq_process.send((AcquisitionProcessSignals.SET_CSR, (key, value)))

    def set_iir_csr(self, *args):
        self.acq_process.send((AcquisitionProcessSignals.SET_IIR_CSR, args))

    def pause_acquisition(self):
        self.acq_process.send((AcquisitionProcessSignals.PAUSE_ACQUISIITON, True))

    def continue_acquisition(self, uuid):
        self.acq_process.send((AcquisitionProcessSignals.CONTINUE_ACQUISITION, uuid))

    def set_raw_acquisition(self, enabled, decimation=None):
        if decimation is None:
            decimation = 0
        self.acq_process.send(
            (AcquisitionProcessSignals.SET_RAW_ACQUISITION, (enabled, decimation))
        )

    def set_dual_channel(self, enabled):
        self.acq_process.send((AcquisitionProcessSignals.SET_DUAL_CHANNEL, enabled))



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import find_packages, setup

version = "0.6.0rc4"

setup(
    name="linien-server",
    version=version,
    author="Benjamin Wiegand",
    author_email="highwaychile@posteo.de",
    maintainer="Bastian Leykauf",
    maintainer_email="leykauf@physik.hu-berlin.de",
    description="Server components of the Linien spectroscopy lock application.",
    long_description="Have a look at the [project repository](https://github.com/linien-org/linien) for installation instructions.",  # noqa: E501
    long_description_content_type="text/markdown",
    url="https://github.com/linien-org/linien",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    entry_points={"console_scripts": ["linien-server=linien_server.server:run_server"]},
    python_requires=">=3.5",
    install_requires=[
        "click>=7.1.2",
        "cma>=3.0.3",
        "myhdl>=0.11",
        "pylpsd>=0.1.4",
        # "pyrp3>=1.1.0", # FIXME: Enable once pyrp3 is on PyPI
        "rpyc>=4.0,<5.0",
        "linien-common=={}".format(version),
    ],
    scripts=[
        "linien_server/linien_start_server.sh",
        "linien_server/linien_stop_server.sh",
        "linien_server/linien_install_requirements.sh",
    ],
    package_data={
        "": [
            "linien.bin",
            "linien_start_server.sh",
            "linien_stop_server.sh",
            "linien_install_requirements.sh",
        ]
    },
)



# Copyright 2014-2015 Robert Jordens <jordens@gmail.com>
# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of redpid.
#
# redpid is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import warnings
from math import ceil, log2, pi

import scipy.signal


def make_filter(name, k=1.0, f=0.0, g=1e20, q=0.5):
    f *= pi

    if name == "LP":  # k/(s + 1)
        b = [f * k / (f + 2), f * k / (f + 2)]
        a = [1, (f - 2) / (f + 2)]

    elif name == "HP":  # k/(1 + 1/s)
        b = [2 * k / (f + 2), 2 * k / (-f - 2)]
        a = [1, (f - 2) / (f + 2)]

    elif name == "AP":  # k*(s - 1)/(s + 1)
        b = [k * (-f + 2) / (f + 2), -k]
        a = [1, (f - 2) / (f + 2)]

    elif name == "I":  # k/s
        b = [f * k / 2, f * k / 2]
        a = [1, -1]

    elif name == "PI":  # k*(s + 1)/(s + 1/g)
        b = [g * k * (f + 2) / (f + 2 * g), g * k * (f - 2) / (f + 2 * g)]
        a = [1, (f - 2 * g) / (f + 2 * g)]

    elif name == "P":  # k
        b = [k]
        a = [1]

    elif name == "PD":  # k*(s + 1)/(1 + s/g)
        b = [g * k * (f + 2) / (f * g + 2), g * k * (f - 2) / (f * g + 2)]
        a = [1, (f * g - 2) / (f * g + 2)]

    elif name == "LP2":  # k/(s**2 + 1 + s/q)
        b = [
            f**2 * k * q / (f**2 * q + 2 * f + 4 * q),
            2 * f**2 * k * q / (f**2 * q + 2 * f + 4 * q),
            f**2 * k * q / (f**2 * q + 2 * f + 4 * q),
        ]
        a = [
            1,
            2 * q * (f**2 - 4) / (f**2 * q + 2 * f + 4 * q),
            (f**2 * q - 2 * f + 4 * q) / (f**2 * q + 2 * f + 4 * q),
        ]

    elif name == "HP2":  # k/(1 + s**(-2) + 1/(q*s))
        b = [
            4 * k * q / (f**2 * q + 2 * f + 4 * q),
            8 * k * q / (-(f**2) * q - 2 * f - 4 * q),
            4 * k * q / (f**2 * q + 2 * f + 4 * q),
        ]
        a = [
            1,
            2 * q * (f**2 - 4) / (f**2 * q + 2 * f + 4 * q),
            (f**2 * q - 2 * f + 4 * q) / (f**2 * q + 2 * f + 4 * q),
        ]

    elif name == "NOTCH":  # k*(s**2 + 1)/(s**2 + 1 + s/q)
        b = [
            k * q * (f**2 + 4) / (f**2 * q + 2 * f + 4 * q),
            2 * k * q * (f**2 - 4) / (f**2 * q + 2 * f + 4 * q),
            k * q * (f**2 + 4) / (f**2 * q + 2 * f + 4 * q),
        ]
        a = [
            1,
            2 * q * (f**2 - 4) / (f**2 * q + 2 * f + 4 * q),
            (f**2 * q - 2 * f + 4 * q) / (f**2 * q + 2 * f + 4 * q),
        ]

    elif name == "IHO":  # 4*k*(s + 1/s + 1/q)/((1 + s/g)*(f*g + 2))
        b = [
            2
            * g
            * k
            * (f**2 * q + 2 * f + 4 * q)
            / (q * (f**2 * g**2 + 4 * f * g + 4)),
            4 * g * k * (f**2 - 4) / (f**2 * g**2 + 4 * f * g + 4),
            2
            * g
            * k
            * (f**2 * q - 2 * f + 4 * q)
            / (q * (f**2 * g**2 + 4 * f * g + 4)),
        ]
        a = [1, 4 / (-f * g - 2), (-f * g + 2) / (f * g + 2)]

    return b, a


def quantize_filter(b, a, shift=None, width=25):
    b, a = [i / a[0] for i in b], [i / a[0] for i in a]

    if shift is None:
        shift = width
        for i in b + a:
            m = ceil(log2(abs(i)))
            if i > 0 and int(m) == m:
                m += 1
            shift = min(shift, int(width - 1 - m))
    s = 1 << shift

    b = [int(round(i * s)) for i in b]
    a = [int(round(i * s)) for i in a]

    m = 1 << (width - 1)
    for i in b + a:
        assert -m <= i < m, (hex(i), hex(m))

    z, p, k = scipy.signal.tf2zpk(b, a)
    if any(abs(_) > 1 for _ in p):
        warnings.warn(
            "unstable filter: z={}, p={}, k={}".format(z, p, k), RuntimeWarning
        )

    return b, a, shift


def get_params(b, a, shift=None, width=25, interval=1):
    b, a, shift = quantize_filter(b, a, shift, width)
    params = {}
    for i, (ai, bi) in enumerate(zip(a, b)):
        params["a%i" % i] = int(-ai)
        params["b%i" % i] = int(bi)
    del params["a0"]
    # params["shift"] = shift
    return b, a, params



csr_constants = {
    "fast_a_iir_c_1_shift": 23,
    "fast_a_iir_c_1_width": 25,
    "fast_a_iir_c_1_interval": 1,
    "fast_a_iir_c_1_latency": 2,
    "fast_a_iir_c_1_order": 1,
    "fast_a_iir_c_1_iterative": 0,
    "fast_a_iir_d_1_shift": 23,
    "fast_a_iir_d_1_width": 25,
    "fast_a_iir_d_1_interval": 1,
    "fast_a_iir_d_1_latency": 3,
    "fast_a_iir_d_1_order": 2,
    "fast_a_iir_d_1_iterative": 0,
    "fast_a_iir_c_2_shift": 23,
    "fast_a_iir_c_2_width": 25,
    "fast_a_iir_c_2_interval": 1,
    "fast_a_iir_c_2_latency": 2,
    "fast_a_iir_c_2_order": 1,
    "fast_a_iir_c_2_iterative": 0,
    "fast_a_iir_d_2_shift": 23,
    "fast_a_iir_d_2_width": 25,
    "fast_a_iir_d_2_interval": 1,
    "fast_a_iir_d_2_latency": 3,
    "fast_a_iir_d_2_order": 2,
    "fast_a_iir_d_2_iterative": 0,
    "fast_b_iir_c_1_shift": 23,
    "fast_b_iir_c_1_width": 25,
    "fast_b_iir_c_1_interval": 1,
    "fast_b_iir_c_1_latency": 2,
    "fast_b_iir_c_1_order": 1,
    "fast_b_iir_c_1_iterative": 0,
    "fast_b_iir_d_1_shift": 23,
    "fast_b_iir_d_1_width": 25,
    "fast_b_iir_d_1_interval": 1,
    "fast_b_iir_d_1_latency": 3,
    "fast_b_iir_d_1_order": 2,
    "fast_b_iir_d_1_iterative": 0,
    "fast_b_iir_c_2_shift": 23,
    "fast_b_iir_c_2_width": 25,
    "fast_b_iir_c_2_interval": 1,
    "fast_b_iir_c_2_latency": 2,
    "fast_b_iir_c_2_order": 1,
    "fast_b_iir_c_2_iterative": 0,
    "fast_b_iir_d_2_shift": 23,
    "fast_b_iir_d_2_width": 25,
    "fast_b_iir_d_2_interval": 1,
    "fast_b_iir_d_2_latency": 3,
    "fast_b_iir_d_2_order": 2,
    "fast_b_iir_d_2_iterative": 0,
    "logic_raw_acquisition_iir_shift": 23,
    "logic_raw_acquisition_iir_width": 25,
    "logic_raw_acquisition_iir_interval": 1,
    "logic_raw_acquisition_iir_latency": 6,
    "logic_raw_acquisition_iir_order": 5,
    "logic_raw_acquisition_iir_iterative": 0,
}

csr = {
    "dna_dna": (28, 0x000, 64, False),
    "fast_a_y_tap": (0, 0x000, 2, True),
    "fast_a_invert": (0, 0x001, 1, True),
    "fast_a_demod_delay": (0, 0x002, 32, True),
    "fast_a_demod_multiplier": (0, 0x006, 4, True),
    "fast_a_x_limit_1_min": (0, 0x007, 25, True),
    "fast_a_x_limit_1_max": (0, 0x00B, 25, True),
    "fast_a_iir_c_1_z0": (0, 0x00F, 27, True),
    "fast_a_iir_c_1_a1": (0, 0x013, 25, True),
    "fast_a_iir_c_1_b0": (0, 0x017, 25, True),
    "fast_a_iir_c_1_b1": (0, 0x01B, 25, True),
    "fast_a_iir_d_1_z0": (0, 0x01F, 27, True),
    "fast_a_iir_d_1_a1": (0, 0x023, 25, True),
    "fast_a_iir_d_1_a2": (0, 0x027, 25, True),
    "fast_a_iir_d_1_b0": (0, 0x02B, 25, True),
    "fast_a_iir_d_1_b1": (0, 0x02F, 25, True),
    "fast_a_iir_d_1_b2": (0, 0x033, 25, True),
    "fast_a_y_limit_1_min": (0, 0x037, 25, True),
    "fast_a_y_limit_1_max": (0, 0x03B, 25, True),
    "fast_a_x_limit_2_min": (0, 0x03F, 25, True),
    "fast_a_x_limit_2_max": (0, 0x043, 25, True),
    "fast_a_iir_c_2_z0": (0, 0x047, 27, True),
    "fast_a_iir_c_2_a1": (0, 0x04B, 25, True),
    "fast_a_iir_c_2_b0": (0, 0x04F, 25, True),
    "fast_a_iir_c_2_b1": (0, 0x053, 25, True),
    "fast_a_iir_d_2_z0": (0, 0x057, 27, True),
    "fast_a_iir_d_2_a1": (0, 0x05B, 25, True),
    "fast_a_iir_d_2_a2": (0, 0x05F, 25, True),
    "fast_a_iir_d_2_b0": (0, 0x063, 25, True),
    "fast_a_iir_d_2_b1": (0, 0x067, 25, True),
    "fast_a_iir_d_2_b2": (0, 0x06B, 25, True),
    "fast_a_y_limit_2_min": (0, 0x06F, 25, True),
    "fast_a_y_limit_2_max": (0, 0x073, 25, True),
    "fast_a_x_clr": (0, 0x077, 1, True),
    "fast_a_x_max": (0, 0x078, 25, False),
    "fast_a_x_min": (0, 0x07C, 25, False),
    "fast_a_out_i_clr": (0, 0x080, 1, True),
    "fast_a_out_i_max": (0, 0x081, 25, False),
    "fast_a_out_i_min": (0, 0x085, 25, False),
    "fast_a_out_q_clr": (0, 0x089, 1, True),
    "fast_a_out_q_max": (0, 0x08A, 25, False),
    "fast_a_out_q_min": (0, 0x08E, 25, False),
    "fast_a_dx_sel": (0, 0x092, 4, True),
    "fast_a_dy_sel": (0, 0x093, 4, True),
    "fast_b_y_tap": (1, 0x000, 2, True),
    "fast_b_invert": (1, 0x001, 1, True),
    "fast_b_demod_delay": (1, 0x002, 32, True),
    "fast_b_demod_multiplier": (1, 0x006, 4, True),
    "fast_b_x_limit_1_min": (1, 0x007, 25, True),
    "fast_b_x_limit_1_max": (1, 0x00B, 25, True),
    "fast_b_iir_c_1_z0": (1, 0x00F, 27, True),
    "fast_b_iir_c_1_a1": (1, 0x013, 25, True),
    "fast_b_iir_c_1_b0": (1, 0x017, 25, True),
    "fast_b_iir_c_1_b1": (1, 0x01B, 25, True),
    "fast_b_iir_d_1_z0": (1, 0x01F, 27, True),
    "fast_b_iir_d_1_a1": (1, 0x023, 25, True),
    "fast_b_iir_d_1_a2": (1, 0x027, 25, True),
    "fast_b_iir_d_1_b0": (1, 0x02B, 25, True),
    "fast_b_iir_d_1_b1": (1, 0x02F, 25, True),
    "fast_b_iir_d_1_b2": (1, 0x033, 25, True),
    "fast_b_y_limit_1_min": (1, 0x037, 25, True),
    "fast_b_y_limit_1_max": (1, 0x03B, 25, True),
    "fast_b_x_limit_2_min": (1, 0x03F, 25, True),
    "fast_b_x_limit_2_max": (1, 0x043, 25, True),
    "fast_b_iir_c_2_z0": (1, 0x047, 27, True),
    "fast_b_iir_c_2_a1": (1, 0x04B, 25, True),
    "fast_b_iir_c_2_b0": (1, 0x04F, 25, True),
    "fast_b_iir_c_2_b1": (1, 0x053, 25, True),
    "fast_b_iir_d_2_z0": (1, 0x057, 27, True),
    "fast_b_iir_d_2_a1": (1, 0x05B, 25, True),
    "fast_b_iir_d_2_a2": (1, 0x05F, 25, True),
    "fast_b_iir_d_2_b0": (1, 0x063, 25, True),
    "fast_b_iir_d_2_b1": (1, 0x067, 25, True),
    "fast_b_iir_d_2_b2": (1, 0x06B, 25, True),
    "fast_b_y_limit_2_min": (1, 0x06F, 25, True),
    "fast_b_y_limit_2_max": (1, 0x073, 25, True),
    "fast_b_x_clr": (1, 0x077, 1, True),
    "fast_b_x_max": (1, 0x078, 25, False),
    "fast_b_x_min": (1, 0x07C, 25, False),
    "fast_b_out_i_clr": (1, 0x080, 1, True),
    "fast_b_out_i_max": (1, 0x081, 25, False),
    "fast_b_out_i_min": (1, 0x085, 25, False),
    "fast_b_out_q_clr": (1, 0x089, 1, True),
    "fast_b_out_q_max": (1, 0x08A, 25, False),
    "fast_b_out_q_min": (1, 0x08E, 25, False),
    "fast_b_dx_sel": (1, 0x092, 4, True),
    "fast_b_dy_sel": (1, 0x093, 4, True),
    "gpio_n_ins": (30, 0x000, 8, False),
    "gpio_n_outs": (30, 0x001, 8, True),
    "gpio_n_oes": (30, 0x002, 8, True),
    "gpio_n_state": (30, 0x003, 14, False),
    "gpio_n_state_clr": (30, 0x005, 1, True),
    "gpio_n_do0_en": (30, 0x006, 14, True),
    "gpio_n_do1_en": (30, 0x008, 14, True),
    "gpio_n_do2_en": (30, 0x00A, 14, True),
    "gpio_n_do3_en": (30, 0x00C, 14, True),
    "gpio_n_do4_en": (30, 0x00E, 14, True),
    "gpio_n_do5_en": (30, 0x010, 14, True),
    "gpio_n_do6_en": (30, 0x012, 14, True),
    "gpio_n_do7_en": (30, 0x014, 14, True),
    "gpio_p_ins": (31, 0x000, 8, False),
    "gpio_p_outs": (31, 0x001, 8, True),
    "gpio_p_oes": (31, 0x002, 8, True),
    "logic_dual_channel": (8, 0x000, 1, True),
    "logic_chain_a_factor": (8, 0x001, 9, True),
    "logic_chain_b_factor": (8, 0x003, 9, True),
    "logic_chain_a_offset": (8, 0x005, 14, True),
    "logic_chain_b_offset": (8, 0x007, 14, True),
    "logic_combined_offset": (8, 0x009, 14, True),
    "logic_out_offset": (8, 0x00B, 14, True),
    "logic_mod_channel": (8, 0x00D, 1, True),
    "logic_control_channel": (8, 0x00E, 1, True),
    "logic_sweep_channel": (8, 0x00F, 2, True),
    "logic_fast_mode": (8, 0x010, 1, True),
    "logic_slow_value": (8, 0x011, 14, False),
    "logic_slow_decimation": (8, 0x013, 5, True),
    "logic_analog_out_1": (8, 0x014, 15, True),
    "logic_analog_out_2": (8, 0x016, 15, True),
    "logic_analog_out_3": (8, 0x018, 15, True),
    "logic_mod_amp": (8, 0x01A, 14, True),
    "logic_mod_freq": (8, 0x01C, 32, True),
    "logic_sweep_step": (8, 0x020, 30, True),
    "logic_sweep_min": (8, 0x024, 14, True),
    "logic_sweep_max": (8, 0x026, 14, True),
    "logic_sweep_run": (8, 0x028, 1, True),
    "logic_sweep_pause": (8, 0x029, 1, True),
    "logic_limit_error_signal_min": (8, 0x02A, 25, True),
    "logic_limit_error_signal_max": (8, 0x02E, 25, True),
    "logic_limit_fast1_min": (8, 0x032, 14, True),
    "logic_limit_fast1_max": (8, 0x034, 14, True),
    "logic_limit_fast2_min": (8, 0x036, 14, True),
    "logic_limit_fast2_max": (8, 0x038, 14, True),
    "logic_pid_setpoint": (8, 0x03A, 25, True),
    "logic_pid_kp": (8, 0x03E, 14, True),
    "logic_pid_ki": (8, 0x040, 14, True),
    "logic_pid_reset": (8, 0x042, 1, True),
    "logic_pid_kd": (8, 0x043, 14, True),
    "logic_autolock_robust_time_scale": (8, 0x045, 14, True),
    "logic_autolock_robust_N_instructions": (8, 0x047, 5, True),
    "logic_autolock_robust_final_wait_time": (8, 0x048, 14, True),
    "logic_autolock_robust_peak_height_0": (8, 0x04A, 28, True),
    "logic_autolock_robust_peak_height_1": (8, 0x04E, 28, True),
    "logic_autolock_robust_peak_height_2": (8, 0x052, 28, True),
    "logic_autolock_robust_peak_height_3": (8, 0x056, 28, True),
    "logic_autolock_robust_peak_height_4": (8, 0x05A, 28, True),
    "logic_autolock_robust_peak_height_5": (8, 0x05E, 28, True),
    "logic_autolock_robust_peak_height_6": (8, 0x062, 28, True),
    "logic_autolock_robust_peak_height_7": (8, 0x066, 28, True),
    "logic_autolock_robust_peak_height_8": (8, 0x06A, 28, True),
    "logic_autolock_robust_peak_height_9": (8, 0x06E, 28, True),
    "logic_autolock_robust_peak_height_10": (8, 0x072, 28, True),
    "logic_autolock_robust_peak_height_11": (8, 0x076, 28, True),
    "logic_autolock_robust_peak_height_12": (8, 0x07A, 28, True),
    "logic_autolock_robust_peak_height_13": (8, 0x07E, 28, True),
    "logic_autolock_robust_peak_height_14": (8, 0x082, 28, True),
    "logic_autolock_robust_peak_height_15": (8, 0x086, 28, True),
    "logic_autolock_robust_peak_height_16": (8, 0x08A, 28, True),
    "logic_autolock_robust_peak_height_17": (8, 0x08E, 28, True),
    "logic_autolock_robust_peak_height_18": (8, 0x092, 28, True),
    "logic_autolock_robust_peak_height_19": (8, 0x096, 28, True),
    "logic_autolock_robust_peak_height_20": (8, 0x09A, 28, True),
    "logic_autolock_robust_peak_height_21": (8, 0x09E, 28, True),
    "logic_autolock_robust_peak_height_22": (8, 0x0A2, 28, True),
    "logic_autolock_robust_peak_height_23": (8, 0x0A6, 28, True),
    "logic_autolock_robust_peak_height_24": (8, 0x0AA, 28, True),
    "logic_autolock_robust_peak_height_25": (8, 0x0AE, 28, True),
    "logic_autolock_robust_peak_height_26": (8, 0x0B2, 28, True),
    "logic_autolock_robust_peak_height_27": (8, 0x0B6, 28, True),
    "logic_autolock_robust_peak_height_28": (8, 0x0BA, 28, True),
    "logic_autolock_robust_peak_height_29": (8, 0x0BE, 28, True),
    "logic_autolock_robust_peak_height_30": (8, 0x0C2, 28, True),
    "logic_autolock_robust_peak_height_31": (8, 0x0C6, 28, True),
    "logic_autolock_robust_wait_for_0": (8, 0x0CA, 14, True),
    "logic_autolock_robust_wait_for_1": (8, 0x0CC, 14, True),
    "logic_autolock_robust_wait_for_2": (8, 0x0CE, 14, True),
    "logic_autolock_robust_wait_for_3": (8, 0x0D0, 14, True),
    "logic_autolock_robust_wait_for_4": (8, 0x0D2, 14, True),
    "logic_autolock_robust_wait_for_5": (8, 0x0D4, 14, True),
    "logic_autolock_robust_wait_for_6": (8, 0x0D6, 14, True),
    "logic_autolock_robust_wait_for_7": (8, 0x0D8, 14, True),
    "logic_autolock_robust_wait_for_8": (8, 0x0DA, 14, True),
    "logic_autolock_robust_wait_for_9": (8, 0x0DC, 14, True),
    "logic_autolock_robust_wait_for_10": (8, 0x0DE, 14, True),
    "logic_autolock_robust_wait_for_11": (8, 0x0E0, 14, True),
    "logic_autolock_robust_wait_for_12": (8, 0x0E2, 14, True),
    "logic_autolock_robust_wait_for_13": (8, 0x0E4, 14, True),
    "logic_autolock_robust_wait_for_14": (8, 0x0E6, 14, True),
    "logic_autolock_robust_wait_for_15": (8, 0x0E8, 14, True),
    "logic_autolock_robust_wait_for_16": (8, 0x0EA, 14, True),
    "logic_autolock_robust_wait_for_17": (8, 0x0EC, 14, True),
    "logic_autolock_robust_wait_for_18": (8, 0x0EE, 14, True),
    "logic_autolock_robust_wait_for_19": (8, 0x0F0, 14, True),
    "logic_autolock_robust_wait_for_20": (8, 0x0F2, 14, True),
    "logic_autolock_robust_wait_for_21": (8, 0x0F4, 14, True),
    "logic_autolock_robust_wait_for_22": (8, 0x0F6, 14, True),
    "logic_autolock_robust_wait_for_23": (8, 0x0F8, 14, True),
    "logic_autolock_robust_wait_for_24": (8, 0x0FA, 14, True),
    "logic_autolock_robust_wait_for_25": (8, 0x0FC, 14, True),
    "logic_autolock_robust_wait_for_26": (8, 0x0FE, 14, True),
    "logic_autolock_robust_wait_for_27": (8, 0x100, 14, True),
    "logic_autolock_robust_wait_for_28": (8, 0x102, 14, True),
    "logic_autolock_robust_wait_for_29": (8, 0x104, 14, True),
    "logic_autolock_robust_wait_for_30": (8, 0x106, 14, True),
    "logic_autolock_robust_wait_for_31": (8, 0x108, 14, True),
    "logic_autolock_fast_target_position": (8, 0x10A, 14, True),
    "logic_autolock_request_lock": (8, 0x10C, 1, True),
    "logic_autolock_autolock_mode": (8, 0x10D, 2, True),
    "logic_autolock_lock_running": (8, 0x10E, 1, False),
    "logic_raw_acquisition_iir_z0": (8, 0x10F, 27, True),
    "logic_raw_acquisition_iir_a1": (8, 0x113, 25, True),
    "logic_raw_acquisition_iir_a2": (8, 0x117, 25, True),
    "logic_raw_acquisition_iir_a3": (8, 0x11B, 25, True),
    "logic_raw_acquisition_iir_a4": (8, 0x11F, 25, True),
    "logic_raw_acquisition_iir_a5": (8, 0x123, 25, True),
    "logic_raw_acquisition_iir_b0": (8, 0x127, 25, True),
    "logic_raw_acquisition_iir_b1": (8, 0x12B, 25, True),
    "logic_raw_acquisition_iir_b2": (8, 0x12F, 25, True),
    "logic_raw_acquisition_iir_b3": (8, 0x133, 25, True),
    "logic_raw_acquisition_iir_b4": (8, 0x137, 25, True),
    "logic_raw_acquisition_iir_b5": (8, 0x13B, 25, True),
    "logic_control_signal_clr": (8, 0x13F, 1, True),
    "logic_control_signal_max": (8, 0x140, 25, False),
    "logic_control_signal_min": (8, 0x144, 25, False),
    "logic_combined_error_signal_clr": (8, 0x148, 1, True),
    "logic_combined_error_signal_max": (8, 0x149, 25, False),
    "logic_combined_error_signal_min": (8, 0x14D, 25, False),
    "logic_combined_error_signal_filtered_clr": (8, 0x151, 1, True),
    "logic_combined_error_signal_filtered_max": (8, 0x152, 25, False),
    "logic_combined_error_signal_filtered_min": (8, 0x156, 25, False),
    "scopegen_external_trigger": (6, 0x000, 1, True),
    "scopegen_dac_a_clr": (6, 0x001, 1, True),
    "scopegen_dac_a_max": (6, 0x002, 25, False),
    "scopegen_dac_a_min": (6, 0x006, 25, False),
    "scopegen_dac_b_clr": (6, 0x00A, 1, True),
    "scopegen_dac_b_max": (6, 0x00B, 25, False),
    "scopegen_dac_b_min": (6, 0x00F, 25, False),
    "scopegen_adc_a_sel": (6, 0x013, 4, True),
    "scopegen_adc_b_sel": (6, 0x014, 4, True),
    "scopegen_adc_a_q_sel": (6, 0x015, 4, True),
    "scopegen_adc_b_q_sel": (6, 0x016, 4, True),
    "slow_pid_setpoint": (2, 0x000, 14, True),
    "slow_pid_kp": (2, 0x002, 14, True),
    "slow_pid_ki": (2, 0x004, 14, True),
    "slow_pid_reset": (2, 0x006, 1, True),
    "slow_pid_kd": (2, 0x007, 14, True),
    "slow_limit_min": (2, 0x009, 14, True),
    "slow_limit_max": (2, 0x00B, 14, True),
    "slow_out_clr": (2, 0x00D, 1, True),
    "slow_out_max": (2, 0x00E, 25, False),
    "slow_out_min": (2, 0x012, 25, False),
    "xadc_temp": (29, 0x000, 12, False),
    "xadc_v": (29, 0x002, 12, False),
    "xadc_a": (29, 0x004, 12, False),
    "xadc_b": (29, 0x006, 12, False),
    "xadc_c": (29, 0x008, 12, False),
    "xadc_d": (29, 0x00A, 12, False),
}
states = [
    "force",
    "di0",
    "di1",
    "di2",
    "di3",
    "di4",
    "di5",
    "di6",
    "di7",
    "robust_watching",
    "robust_turn_on_lock",
    "robust_sign_equal",
    "robust_over_threshold",
    "robust_waited_long_enough",
]
signals = [
    "zero",
    "fast_a_x",
    "fast_a_out_i",
    "fast_a_out_q",
    "fast_b_x",
    "fast_b_out_i",
    "fast_b_out_q",
    "slow_out",
    "scopegen_dac_a",
    "scopegen_dac_b",
    "logic_control_signal",
    "logic_combined_error_signal",
    "logic_combined_error_signal_filtered",
]
